from __future__ import annotations

from aperturedb.Entities import Entities


class BoundingBoxes(Entities):
    """
    **The object mapper representation of bounding boxes in ApertureDB.**

    This class is a layer on top of the native query.
    It facilitates interactions with bounding boxes in the database in the pythonic way.
    """
    db_object = "_BoundingBox"



import pytest
import numpy as np
import pandas as pd
import os.path as osp
from aperturedb.Query import QueryBuilder, Query
from aperturedb.Entities import Entities
from aperturedb.Constraints import Constraints
from aperturedb.Images import Images
from aperturedb.Utils import Utils

import logging
logger = logging.getLogger(__name__)


@pytest.mark.slow
class TestEntityLoader():
    def assertEqual(self, expected, actual):
        if expected != actual:
            raise AssertionError(
                "Expected {}, got {}".format(expected, actual))

    def assertTrue(self, condition):
        if not condition:
            raise AssertionError("Condition not true")

    def test_Loader(self, utils: Utils, insert_data_from_csv):
        classes = ["_Image", "_Descriptor", "Person", "_BoundingBox"]
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())

        results = [utils.create_entity_index(
            e, "id") for e in classes]
        self.assertTrue(all(results))

        # Insert Person Nodes
        data, _ = insert_data_from_csv("./input/persons.adb.csv")
        self.assertEqual(len(data), utils.count_entities("Person"))

        # Insert Images
        data, _ = insert_data_from_csv(in_csv_file = "./input/images.adb.csv")
        self.assertEqual(len(data), utils.count_images())

        # Insert Connections
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/connections-persons-images.adb.csv")
        self.assertEqual(
            len(data), utils.count_connections("has_image"))

        # Insert BBoxes
        data, _ = insert_data_from_csv(in_csv_file="./input/bboxes.adb.csv")
        self.assertEqual(len(data), utils.count_bboxes())

    def test_LoaderDescriptorsImages(self, utils, insert_data_from_csv):
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())

        # Insert Images, images may be there already, and that case should
        # be handled correctly by contraints
        data, _ = insert_data_from_csv(in_csv_file = "./input/images.adb.csv")
        self.assertEqual(len(data), utils.count_images())

        # Insert DescriptorsSet
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/descriptorset.adb.csv")
        self.assertEqual(len(data), utils.count_entities("_DescriptorSet"))

        # Insert Descriptors

        sets = ["setA", "setB"]

        total_descriptors = 0
        for setname in sets:
            data, _ = insert_data_from_csv(
                in_csv_file = "./input/" + setname + ".adb.csv")

            total_descriptors += len(data)
            self.assertEqual(total_descriptors,
                             utils.count_entities("_Descriptor"))

    def test_BlobLoader(self, db, utils, insert_data_from_csv):
        # Assert that we have a clean slate to begin with.
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())

        data, _ = insert_data_from_csv(in_csv_file = "./input/blobs.adb.csv")
        self.assertEqual(len(data), utils.count_entities("_Blob"))

        query = [{
            "FindBlob": {
                "blobs": True,
                "results": {
                    "list": ["license"],
                    "limit": 1
                }
            }
        }]

        res, blob = db.query(query)
        self.assertEqual(len(blob), 1)

        arr = np.frombuffer(blob[0])
        self.assertEqual(arr[2], 3.3)

    def test_conditional_add(self, utils, insert_data_from_csv):
        logger.debug(f"Cleaning existing data")
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        # insert one of the images from image csv, for bboxes to refer to.
        images, _ = insert_data_from_csv(
            in_csv_file="./input/images.adb.csv", rec_count=1)
        self.assertEqual(len(images), 1)

        # Insert BBoxes with repeated entries.
        # There is just one unique entry in the input csv.
        logger.debug(f"Inserting bounding box data")
        boxes, _ = insert_data_from_csv(
            in_csv_file="./input/bboxes-constraints.adb.csv")
        self.assertEqual(3, len(boxes))
        self.assertEqual(1, utils.count_bboxes())

    def test_updateif_fails(self, db, utils, modify_data_from_csv):
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        # verifies loading with empty database.
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/persons-update.adb.csv")
        self.assertEqual(len(data), utils.count_entities("Person"))
        # verifies updateif will not update if criteria doesn't pass.
        update_data, _ = modify_data_from_csv(
            in_csv_file = "./input/persons-update-oldversion.adb.csv")
        self.assertEqual(len(data), utils.count_entities("Person"))
        all_persons = Entities.retrieve(db,
                                        spec=Query.spec(with_class="Person"))
        # if updated, age will be above 200.
        self.assertEqual(
            len(list(filter(lambda p: p['version_id'] == 2, all_persons))), 0)
        self.assertEqual(
            len(list(filter(lambda p: p['age'] >= 200, all_persons))), 0)

    def test_updateif_passes(self, db, utils, modify_data_from_csv):
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/persons-update.adb.csv")
        self.assertEqual(len(data), utils.count_entities("Person"))
        update_data, _ = modify_data_from_csv(
            in_csv_file = "./input/persons-update-newversion.adb.csv")
        self.assertEqual(len(data), utils.count_entities("Person"))
        all_persons = Entities.retrieve(db,
                                        spec=Query.spec(with_class="Person"))
        # if updated, age will be above 200.
        self.assertEqual(
            len(list(filter(lambda p: p['version_id'] == 2, all_persons))), len(update_data))
        self.assertEqual(
            len(list(filter(lambda p: p['age'] >= 200, all_persons))), len(update_data))

    # Test updating with conditional for > and only some records being updated
    def test_updateif_partial_age(self, db, utils, modify_data_from_csv):
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/persons-update.adb.csv")
        self.assertEqual(len(data), utils.count_entities("Person"))

        all_persons = Entities.retrieve(db,
                                        spec=Query.spec(with_class="Person"))
        previous_old_age = len(
            list(filter(lambda p: p['age'] > 30, all_persons)))
        update_data, _ = modify_data_from_csv(
            in_csv_file = "./input/persons-update-olderage.adb.csv")
        all_persons = Entities.retrieve(db,
                                        spec=Query.spec(with_class="Person"))
        # 200 should be addded to all that were above 30, so both should match.
        new_old_age = len(list(filter(lambda p: p['age'] > 200, all_persons)))
        self.assertEqual(previous_old_age, new_old_age)

    # ensure blob adding works.
    def test_updateif_add_image(self, db, utils, modify_data_from_csv):
        # load images
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_updateif_baseload.adb.csv")
        self.assertEqual(len(data), utils.count_images())
        # load images with partial existance
        update_data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_updateif_mixednew.adb.csv")
        self.assertEqual(len(update_data), utils.count_images())

    def get_first_image_size(self, file):
        size = 0
        df = pd.read_csv(file)
        csv_dir = osp.dirname(file)
        with open(osp.join(csv_dir, df.loc[0, 'filename']), "rb") as f:
            data = f.read()
            size = len(data)
        return size
    # Test that images are immutable.

    def test_updateif_image_immutable(self, db, utils, modify_data_from_csv):
        # load images

        big_img_size = self.get_first_image_size(
            "./input/images_updateif_fail_updates.adb.csv")

        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_updateif_fail_baseload.adb.csv")
        self.assertEqual(len(data), utils.count_images())
        # load images with partial existance
        update_data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_updateif_fail_updates.adb.csv")
        # images in the second csv are much bigger than from the first.

        images = Images.retrieve(db,
                                 spec=Query.spec(constraints=Constraints().greaterequal("version_id", 2)))
        images.search()
        # all db images must be less than half the size of a larger changed image.
        # filter will include any larger, we expect 0.
        self.assertEqual(0, len(list(
            filter(lambda imglen: imglen > big_img_size / 2, [len(images.get_image_by_index(
                db_img_id)) for db_img_id in range(len(images.images_ids))])
        )))
        # all images were updated - images with version_id >= 2 is all images in db.
        self.assertEqual(len(images), utils.count_images())
        return

    # ensure ImageForceNewestCSV can create images and then only create the additional ones in mixednew.
    def test_imageforce_load_base(self, db, utils, modify_data_from_csv):
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_forceupdate_baseload.adb.csv")
        self.assertEqual(len(data), utils.count_images())
        update_data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_forceupdate_mixednew.adb.csv")
        self.assertEqual(len(update_data), utils.count_images())

    # ensure ImageForceNewestCSV does not update images when not given a way to see if an image has changed.

    def test_imageforce_load_nonupdate(self, db, utils, modify_data_from_csv):
        big_img_size = self.get_first_image_size(
            "./input/images_updateif_fail_updates.adb.csv")
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_forceupdate_fail_base.adb.csv")
        self.assertEqual(len(data), utils.count_images())
        update_data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_forceupdate_fail_updates.adb.csv")
        images = Images.retrieve(db,
                                 spec=Query.spec(constraints=Constraints().greaterequal("version_id", 2)))
        images.search()
        # all db images must be less than half the size of a larger changed image.
        # filter will include any larger, we expect 0.
        self.assertEqual(0, len(list(
            filter(lambda imglen: imglen > big_img_size / 2, [len(images.get_image_by_index(
                db_img_id)) for db_img_id in range(len(images.images_ids))])
        )))
        # all images were updated
        self.assertEqual(len(images), utils.count_images())

    def test_imageforce_update(self, db, utils, modify_data_from_csv):
        big_img_size = self.get_first_image_size(
            "./input/images_updateif_fail_updates.adb.csv")
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_forceupdate_blob_baseload.adb.csv")
        self.assertEqual(len(data), utils.count_images())
        update_data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_forceupdate_updates.adb.csv")
        images = Images.retrieve(db,
                                 spec=Query.spec(constraints=Constraints().greaterequal("version_id", 2)))
        images.search()
        # all images were updated
        self.assertEqual(len(update_data), len(list(
            filter(lambda imglen: imglen > big_img_size / 2, [len(images.get_image_by_index(
                db_img_id)) for db_img_id in range(len(images.images_ids))])
        )))
        self.assertEqual(len(images), utils.count_images())

    def test_image_sparse(self, db, utils, modify_data_from_csv):
        self.assertTrue(utils.remove_all_indexes())
        self.assertTrue(utils.remove_all_objects())
        data, _ = modify_data_from_csv(
            in_csv_file = "./input/images_sparseload_base.adb.csv")

        self.assertEqual(len(data), utils.count_images())
        fulldata, _ = modify_data_from_csv(
            in_csv_file = "./input/images_sparseload_full.adb.csv")
        self.assertEqual(len(data) * 2, utils.count_images())


# test EntityUpdateCSV with Person entity.


class TestUpdatePersonEntityCSV():
    def assertEqual(self, expected, actual):
        if expected != actual:
            raise AssertionError(
                "Expected {}, got {}".format(expected, actual))

    def assertTrue(self, condition):
        if not condition:
            raise AssertionError("Condition not true")
    # verifies adding and no update when version isn't greater.


# Test functionality of ImageUpdateCSV loader.
@pytest.mark.skip(reason="Old csv not generated")
class TestImageUpdateCSV():
    def assertEqual(self, expected, actual):
        if expected != actual:
            raise AssertionError(
                "Expected {}, got {}".format(expected, actual))

    def assertTrue(self, condition):
        if not condition:
            raise AssertionError("Condition not true")

    def test_images(self, utils, modify_data_from_csv):
        data = modify_data_from_csv(
            in_csv_file= "./input/images_update_and_add.adb.csv")
        self.assertEqual(True, True)  # how to verify.

# Test functionality of ImageForceNewestCSV loader.


@pytest.mark.skip(reason="Old csv not generated")
class TestImageForceNewestCSV():
    def assertEqual(self, expected, actual):
        if expected != actual:
            raise AssertionError(
                "Expected {}, got {}".format(expected, actual))

    def assertTrue(self, condition):
        if not condition:
            raise AssertionError("Condition not true")

    def test_images(self, utils, modify_data_from_csv):
        #
        data = modify_data_from_csv(
            in_csv_file= "./input/images_newest_blobs.adb.csv")
        self.assertEqual(True, True)  # how to verify.


@pytest.mark.external_network
@pytest.mark.remote_credentials
class TestURILoader():
    def assertEqual(self, expected, actual):
        if expected != actual:
            raise AssertionError(
                "Expected {}, got {}".format(expected, actual))

    def test_S3ImageLoader(self, utils, insert_data_from_csv):
        self.assertEqual(utils.remove_all_indexes(), True)
        self.assertEqual(utils.remove_all_objects(), True)
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/s3_images.adb.csv")
        self.assertEqual(len(data), utils.count_images())

    def test_HttpImageLoader(self, utils, insert_data_from_csv):
        self.assertEqual(utils.remove_all_indexes(), True)
        self.assertEqual(utils.remove_all_objects(), True)
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/http_images.adb.csv")
        self.assertEqual(len(data), utils.count_images())

    def test_GSImageLoader(self, utils, insert_data_from_csv):
        self.assertEqual(utils.remove_all_indexes(), True)
        self.assertEqual(utils.remove_all_objects(), True)
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/gs_images.adb.csv")
        self.assertEqual(len(data), utils.count_images())

    def test_S3VideoLoader(self, utils, insert_data_from_csv):
        self.assertEqual(utils.remove_all_indexes(), True)
        self.assertEqual(utils.remove_all_objects(), True)
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/s3_videos.adb.csv")
        self.assertEqual(len(data), utils.count_entities("_Video"))

    def test_HttpVideoLoader(self, utils, insert_data_from_csv):
        self.assertEqual(utils.remove_all_indexes(), True)
        self.assertEqual(utils.remove_all_objects(), True)
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/http_videos.adb.csv")
        self.assertEqual(len(data), utils.count_entities("_Video"))

    def test_GSVideoLoader(self, utils, insert_data_from_csv):
        self.assertEqual(utils.remove_all_indexes(), True)
        self.assertEqual(utils.remove_all_objects(), True)
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/gs_videos.adb.csv")
        self.assertEqual(len(data), utils.count_entities("_Video"))



"""
**These are miscellaneous helper methods to render responses from ApertureDB in a jupyter environment.**
"""

import os
from typing import List
import cv2
import numpy as np

from PIL import Image
from IPython.display import Video
from IPython.display import display as ds
from base64 import b64encode
import matplotlib.pyplot as plt


DESTINATION_FOLDER = "results"
BOX_TEMPORAL_WINDOW = 10


class Rectangle:
    def __init__(self, x: int = 0, y: int = 0, width: int = 0, height: int = 0):
        self.x = x
        self.y = y
        self.width = width
        self.height = height


class BoundingBox:
    """
    **A class which combines a box and a label.**
    """

    def __init__(self, r: Rectangle = None, label: str = None) -> None:
        self.r = r
        self.label = label


class TemporalBoundingBox:
    """
    **A class that represents a box, but also associated time in the range of frame start and end**
    """

    def __init__(self, bb: BoundingBox = None, start_frame: int = 0, end_frame = 0) -> None:
        self.bb = bb
        self.start_frame = start_frame
        self.end_frame = end_frame


def check_folder(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)


def display(images_array, save=False):
    """
    Show images with matplotlib
    :::note
    This method was used by ApertureDB to display images, but it is not recommended anymore.
    It will not draw annotation.

    Instead, when using with JSON queries, make a instance of the Images class and call the display method.
    :

    ```python
    from aperturedb.Images import Images
    result, response, blobs = execute_batch([{"FindImage":{"uniqueids": True}}], [], db)
    wrapper = Images(db, response=response[0]["FindImage"]["entities"])
    wrapper.display()
    ```
    """

    for im in images_array:
        nparr = np.fromstring(im, np.uint8)
        image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        plt.subplots()
        plt.imshow(image)

    if save:
        check_folder(DESTINATION_FOLDER)
        counter = 0
        for im in images_array:
            img_file = DESTINATION_FOLDER + '/res_' + str(counter) + '.jpg'
            counter += 1
            fd = open(img_file, 'wb')
            fd.write(im)
            fd.close()


def draw_bboxes(image, boxes=[], tags=[], save=False):

    nparr    = np.fromstring(image, np.uint8)
    cv_image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

    # Draw a rectangle around the faces
    counter = 0
    for coords in boxes:
        left   = coords["x"]
        top    = coords["y"]
        right  = coords["x"] + coords["width"]
        bottom = coords["y"] + coords["height"]
        cv2.rectangle(cv_image, (left, top), (right, bottom), (0, 255, 0), 2)
        y = top - 15 if top - 15 > 15 else top + 15
        cv2.putText(cv_image, tags[counter], (left, y),
                    cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 255, 0), 2)
        counter += 1

    cv_image_rgb = cv2.cvtColor(cv_image, cv2.COLOR_BGR2RGB)
    ds(Image.fromarray(cv_image_rgb))

    if save:
        check_folder(DESTINATION_FOLDER)
        img_file = DESTINATION_FOLDER + '/res_bboxes.jpg'
        cv2.imwrite(img_file, cv_image)


def save_video(blob, name):
    check_folder(DESTINATION_FOLDER)
    with open(name, 'wb') as fd:
        fd.write(blob)


def display_video_mp4(blob):
    """
    **Display a video using IPython.display**

    Args:
        blob (bytearray): A blob that is the Video.
    """
    name = DESTINATION_FOLDER + "/" + "video_tmp.mp4"
    save_video(blob, name)
    ds(Video(name, embed=True))


def annotate_video(blob, bboxes: List[TemporalBoundingBox] = []):
    """
    **Place annotations on a video framewise**

    Args:
        blob (bytearray): Video blob returned from the database
        bboxes (List[TemporalBoundingBox], optional): List of boxes to be drawn. Defaults to [].
    """
    name = DESTINATION_FOLDER + "/" + "video_tmp.mp4"
    save_video(blob, name)
    cap = cv2.VideoCapture(name)
    if cap.isOpened():
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        res = (int(width), int(height))
        fourcc = cv2.VideoWriter_fourcc(*'avc1')  # codec

        out = cv2.VideoWriter(os.path.join(
            DESTINATION_FOLDER, 'output.mp4'), fourcc, cap.get(cv2.CAP_PROP_FPS), res)
        frame = None
        count = 0
        while True:
            try:
                is_success, frame = cap.read()
                filtered = list(filter(lambda x: abs(
                    x.start_frame - count) < BOX_TEMPORAL_WINDOW or
                    abs(x.end_frame - count) < BOX_TEMPORAL_WINDOW, bboxes))
                if len(filtered) > 0:
                    start = (filtered[0].bb.r.x, filtered[0].bb.r.y)
                    end = (filtered[0].bb.r.x + filtered[0].bb.r.width,
                           filtered[0].bb.r.y + filtered[0].bb.r.height)
                    color = (255, 255, 255)
                    thickness = 2
                    annotated = cv2.rectangle(
                        frame, start, end, color, thickness)
                else:
                    annotated = frame
            except cv2.error as e:
                print(e)
                break

            if not is_success:
                break

            out.write(annotated)
            count += 1

        out.release()
    else:
        print("Unable to open cap")


def display_annotated_video(blob, bboxes: List[TemporalBoundingBox] = []):
    """
    Returns a HTML representation with a column filled
    with video entities.
    """
    annotate_video(blob, bboxes)
    video_path = os.path.join(DESTINATION_FOLDER, 'output.mp4')
    mp4 = open(video_path, "rb").read()
    data_url = "data:video/mp4;base64," + b64encode(mp4).decode()
    return f"""<div style='max-width: 100%; overflow: auto;'><video width=400 controls>
        <source src="{data_url}" type="video/mp4">
    </video></div>"""



from __future__ import annotations
from enum import Enum


class Conjunction(Enum):
    AND = "all"
    OR = "any"


class Constraints(object):
    """
    **Constraints object for the Object mapper API**
    """

    def __init__(self, conjunction: Conjunction = Conjunction.AND):
        self._conjunction = conjunction.value
        self.constraints = {
            conjunction.value: {
            }
        }

    def equal(self, key, value) -> Constraints:
        self.constraints[self._conjunction][key] = ["==", value]
        return self

    def notequal(self, key, value) -> Constraints:
        self.constraints[self._conjunction][key] = ["!=", value]
        return self

    def greaterequal(self, key, value) -> Constraints:
        self.constraints[self._conjunction][key] = [">=", value]
        return self

    def greater(self, key, value) -> Constraints:
        self.constraints[self._conjunction][key] = [">", value]
        return self

    def lessequal(self, key, value) -> Constraints:
        self.constraints[self._conjunction][key] = ["<=", value]
        return self

    def less(self, key, value) -> Constraints:
        self.constraints[self._conjunction][key] = ["<", value]
        return self

    def is_in(self, key, val_array) -> Constraints:
        self.constraints[self._conjunction][key] = ["in", val_array]
        return self

    def check(self, entity):
        for key, op in self.constraints.items():
            if key not in entity:
                return False
            if op[0] == "==":
                if not entity[key] == op[1]:
                    return False
            elif op[0] == ">=":
                if not entity[key] >= op[1]:
                    return False
            elif op[0] == ">":
                if not entity[key] > op[1]:
                    return False
            elif op[0] == "<=":
                if not entity[key] <= op[1]:
                    return False
            elif op[0] == "<":
                if not entity[key] < op[1]:
                    return False
            elif op[0] == "in":
                if not entity[key] in op[1]:
                    return False
            else:
                raise Exception("invalid constraint operation: " + op[0])
        return True



from aperturedb.Subscriptable import Subscriptable
import logging

logger = logging.getLogger(__name__)


class Transformer(Subscriptable):
    """
    Transformer is an abstract class that can be used to transform
    data before ingestion into aperturedb.

    :::info
    **Some build in transformers:**
        - CommonProperties: Add common properties to the data
        - ImageProperties: Add image properties to the data
        - Facenet: Add facenet embeddings to the data
    :::


    [Example](https://github.com/aperture-data/aperturedb-python/blob/develop/examples/similarity_search/add_faces.py) of how to use transformers:
        ```python
        from CelebADataKaggle import CelebADataKaggle
        from aperturedb.transformers.facenet_pytorch_embeddings import FacenetPyTorchEmbeddings
        from aperturedb.transformers.common_properties import CommonProperties
        from aperturedb.transformers.image_properties import ImageProperties

        .
        .
        .

        dataset = CelebADataKaggle()

        # Here's a pipeline that adds extra properties to the celebA dataset
        dataset = CommonProperties(
            dataset,
            adb_data_source="kaggle-celebA",
            adb_main_object="Face",
            adb_annoted=True)

        # some useful properties for the images
        dataset = ImageProperties(dataset)

        # Add the embeddings generated through facenet.
        dataset = FacenetPyTorchEmbeddings(dataset)

        ```

    """

    def __init__(self, data: Subscriptable, **kwargs) -> None:
        self.data = data

        # Inspect the first element to get the number of queries and blobs
        x = self.data[0]
        self._queries = len(x[0])
        self._blobs = len(x[1])
        self._blob_index = []
        self._add_image_index = []

        bc = 0
        for i, c in enumerate(x[0]):
            command = list(c.keys())[0]
            if command in ["AddImage", "AddDescriptor", "AddVideo", "AddBlob"]:
                self._blob_index.append(i)
                if command == "AddImage":
                    self._add_image_index.append(i)
                bc += 1
        logger.info(f"Found {bc} blobs in the data")
        logger.info(
            f"Found {len(self._add_image_index)} AddImage commands in the data")

        self.ncalls = 0
        self.cumulative_time = 0

    def getitem(self, subscript):
        raise NotImplementedError("Needs to be subclassed")

    def __len__(self):
        return len(self.data)
